/*
                            Print Options & Dialog
 */

#ifndef PRINTOPTS_H
#define PRINTOPTS_H

#if !defined(WIN32)

#include <gtk/gtk.h>
#include "imgview.h"


typedef struct _print_values_struct	print_values_struct;
#define PRINT_VALUES(p)		((print_values_struct *)(p))

typedef struct _print_dlg_struct	print_dlg_struct;
#define PRINT_DLG(p)			((print_dlg_struct *)(p))


/*
 *	Output:
 */
typedef enum {
	PRINT_OUTPUT_TO_PRINTER,
	PRINT_OUTPUT_TO_FILE
} print_output;

/*
 *	Visual:
 */
typedef enum {
	PRINT_VISUAL_BW,		/* Black & White */
	PRINT_VISUAL_GREYSCALE,		/* Greyscale */
	PRINT_VISUAL_RGB		/* Color */
} print_visual;

/*
 *	Orientation:
 */
typedef enum {
	PRINT_ORIENTATION_PORTRAIT,
	PRINT_ORIENTATION_LANDSCAPE
} print_orientation;

/*
 *	Paper Sizes List:
 *
 *	Each entry contains four values:
 *
 *	char *name, char *dimensions, int width, int height, int units
 *
 *	Where width and height are in pixels at 100 dpi.
 */
#define PAPER_SIZE_LIST		{ \
{	"Letter",	"8.5 x 11 in",		850,	1100	}, \
{	"Legal",	"8.5 x 14 in",		850,	1400	}, \
{	"Tabloid",	"11 x 17 in",		1100,	1700	}, \
{	"Ledger",	"17 x 11 in",		1700,	1100	}, \
{	"Executive",	"7.5 x 10 in",		750,	1000	}, \
{	"A3",		"11.69 x 16.53 in",	1169,	1653	}, \
{	"A4",		"8.26 x 11.69 in",	826,	1169	}, \
{	"A5",		"5.83 x 8.26 in", 	583,	826	}, \
{	"B4",		"9.83 x 13.93 in",	983,	1393	}, \
{	"B5",		"7.17 x 10.12 in",	717,	1012	}, \
{	NULL,		NULL,			0,	0	} \
}


/*
 *	Print Values:
 */
struct _print_values_struct {

	/* Flags that indicate which members are defined */
#define PRINT_VALUE_OUTPUT_TYPE		(1 << 0)
#define PRINT_VALUE_PRINT_COMMAND	(1 << 1)
#define PRINT_VALUE_FILE_NAME		(1 << 2)
#define PRINT_VALUE_BRIGHTNESS		(1 << 3)
#define PRINT_VALUE_CONTRAST		(1 << 4)
#define PRINT_VALUE_VISUAL		(1 << 5)
#define PRINT_VALUE_ORIENTATION		(1 << 6)
#define PRINT_VALUE_DPI_X		(1 << 7)
#define PRINT_VALUE_DPI_Y		(1 << 8)
#define PRINT_VALUE_PAPER_WIDTH		(1 << 9)
#define PRINT_VALUE_PAPER_HEIGHT	(1 << 10)
#define PRINT_VALUE_OUTPUT_X		(1 << 11)
#define PRINT_VALUE_OUTPUT_Y		(1 << 12)
#define PRINT_VALUE_OUTPUT_WIDTH	(1 << 13)
#define PRINT_VALUE_OUTPUT_HEIGHT	(1 << 14)
#define PRINT_VALUE_OUTPUT_MAINTAIN_ASPECT	(1 << 15)
	guint		flags;

	print_output	output;
	gchar		*print_command;	/* Shared, can be NULL */
	gchar		*file_name;	/* Shared, can be NULL */
	gfloat		brightness;	/* -1.0 to 1.0 */
	gfloat		contrast;	/* -1.0 to 1.0 */
	print_visual	visual;
	print_orientation orientation;
	gint		dpi_x, dpi_y;
	gint		paper_width,	/* In pixels without dpi applied */
			paper_height;

	/* Output geometry
	 *
	 * Matches upper-left portrait orientation, size will already be
	 * modified for correct orientation
	 */
	gint		output_x,	/* In pixels from upper-left corner */
			output_y;
	gint		output_width,	/* In pixels */
			output_height;
	gboolean	output_maintain_aspect;
};


/*
 *	Print Dialog:
 */
struct _print_dlg_struct {

	gboolean	initialized,
			map_state;
	GtkAccelGroup	*accelgrp;
	GtkWidget	*toplevel;
	gint		freeze_count,
			busy_count;

	/* Cursors */
	GdkCursor	*busy_cur,
			*translate_cur,
			*size_cur;

	/* Source image data that user wants to print */
	imgview_image_struct	*src_img;
	gint			src_frame_num;

	/* Output radios */
	GtkWidget	*output_printer_radio,
			*output_file_radio;
	/* Output entries */
	GtkWidget	*output_printer_entry,	/* Print command */
			*output_file_entry;	/* File name */
	GtkWidget	*output_file_browse_btn;

	/* Visual scales */
	GtkWidget	*brightness_scale;
	GtkAdjustment	*brightness_adj;
	GtkWidget	*contrast_scale;
	GtkAdjustment	*contrast_adj;
	/* Visual radios */
	GtkWidget	*visual_bw_radio,
			*visual_greyscale_radio,
			*visual_color_radio;

	/* Orientation radios */
	GtkWidget	*orientation_portrait_radio,
			*orientation_landscape_radio;

	/* DPI entries, axis relative to paper (not image) */
	GtkWidget	*dpi_x_entry,
			*dpi_y_entry;

	/* Paper size clist */
	GtkWidget	*paper_size_clist;


	/* Output geometry entries */
	GtkWidget	*output_x_spin,
			*output_y_spin,
			*output_width_spin,
			*output_height_spin,
			*output_maintain_aspect_check;

	/* Preview drawing area */
	GtkWidget	*preview_da;

	/* Graphic contexts for the preview drawnig area */
	GdkGC		*gc_base,
			*gc_invert;

	/* Preview ImgView image
	 *
	 * This image is post processed with the modified window
	 * coordinate size, color, brightness, rotations, etc which
	 * makes it suitable and ready to be drawn on the preview_da
	 */
	imgview_image_struct	*preview_img;

	/* Original Preview ImgView image
	 *
	 * It matches the size of the preview_img but does not have
	 * brightness or contrast modifications made to it, this is so
	 * that it can be quickly copied to the preview_img with
	 * modifications to the contrast and brightness
	 */
	imgview_image_struct	*orig_preview_img;

	/* Preview menu */
	GtkWidget	*menu;

	/* Image output geometry (not allocated size of preview image), in
	 * pixels. Relative to upper left corner and assumed orientation
	 * to be PRINT_ORIENTATION_PORTRAIT
	 *
	 * When orientation is set to PRINT_ORIENTATION_LANDSCAPE the
	 * geometry values still represent their values as if the
	 * orientation was PRINT_ORIENTATION_PORTRAIT
	 */
	gint		output_x, output_y;
	gint		output_width, output_height;
	gboolean	output_maintain_aspect;

	guint8		bg_color[4];	/* 4 bytes RGBA */

	/* Buttons */
	GtkWidget	*print_btn,
			*cancel_btn;


	/* Keyboard modifiers mask */
#define PRINT_MODIFIER_ALT	(1 << 0)
#define PRINT_MODIFIER_CTRL	(1 << 1)
#define PRINT_MODIFIER_SHIFT	(1 << 2)
	guint		modifiers;

	/* Preview drag mode */
#define PRINT_DRAG_NONE		0
#define PRINT_DRAG_TRANSLATE	1
#define PRINT_DRAG_SIZE		2
	gint		drag_mode;
	gint		drag_last_x, drag_last_y;

	/* Rectangular drag outline in window coordinates, for drawing
	 * a rectangular rubber band on the preview widget while dragging
	 */
	gboolean	only_draw_outline;	/* Set by PrintDlgPreviewEventCB() and
						 * cleared by PrintDlgPreviewExposeCB()
						 */
	gint		drag_outline_x[2],
			drag_outline_y[2];
	gboolean	have_prev_outline;
	gint		prev_outline_x[2],
			prev_outline_y[2];

};


/* Private (do not use publically) */
extern print_dlg_struct *PrintDlgNew(void);
extern void PrintDlgMap(print_dlg_struct *pw);
extern void PrintDlgUnmap(print_dlg_struct *pw);
extern void PrintDlgDelete(print_dlg_struct *pw);


/* Public */
extern imgview_image_struct *PrintDlgCreateImageFromValues(
	imgview_image_struct *img,
	gint frame_num,
	const print_values_struct *v,
	const guint8 *bg_color			/* 4 bytes RGBA */
);

extern gboolean PrintDlgGetResponse(
	imgview_image_struct *img,		/* Image to print */
	gint frame_num,
	print_values_struct *v,
	const guint8 *bg_color,			/* 4 bytes RGBA */
	GtkWidget *toplevel
);


#endif	/* !WIN32 */

#endif	/* PRINTOPTS_H */
